import { type AuthProvider } from './auth/providers.js';
export { Client as TDF3Client } from '../tdf3/src/client/index.js';
import Header from './nanotdf/models/Header.js';
import { type Source } from './seekable.js';
import { AssertionConfig, AssertionVerificationKeys } from '../tdf3/src/assertions.js';
import { type KasPublicKeyAlgorithm, isPublicKeyAlgorithm } from './access.js';
import { type Manifest } from '../tdf3/src/models/manifest.js';
export { type KasPublicKeyAlgorithm, isPublicKeyAlgorithm };
export type Keys = {
    [keyID: string]: CryptoKey | CryptoKeyPair;
};
export type CreateOptions = {
    autoconfigure?: boolean;
    attributes?: string[];
    byteLimit?: number;
    defaultKASEndpoint?: string;
    signers?: Keys;
    source: Source;
};
export type CreateNanoTDFOptions = CreateOptions & {
    bindingType?: 'ecdsa' | 'gmac';
    ecdsaBindingKeyID?: string;
    signingKeyID?: string;
};
export type CreateNanoTDFCollectionOptions = CreateNanoTDFOptions & {
    maxKeyIterations?: number;
};
export type Metadata = object;
export type MimeType = `${string}/${string}`;
export type SplitStep = {
    kas: string;
    sid?: string;
};
export type CreateZTDFOptions = CreateOptions & {
    assertionConfigs?: AssertionConfig[];
    metadata?: Metadata;
    mimeType?: MimeType;
    splitPlan?: SplitStep[];
    windowSize?: number;
    wrappingKeyAlgorithm?: KasPublicKeyAlgorithm;
};
export type ReadOptions = {
    source: Source;
    allowedKASEndpoints?: string[];
    ignoreAllowlist?: boolean;
    assertionVerificationKeys?: AssertionVerificationKeys;
    noVerify?: boolean;
    concurrencyLimit?: number;
    wrappingKeyAlgorithm?: KasPublicKeyAlgorithm;
};
export type OpenTDFOptions = {
    policyEndpoint?: string;
    authProvider: AuthProvider;
    defaultCreateOptions?: Omit<CreateOptions, 'source'>;
    defaultReadOptions?: Omit<ReadOptions, 'source'>;
    disableDPoP?: boolean;
    dpopKeys?: Promise<CryptoKeyPair>;
    rewrapCacheOptions?: RewrapCacheOptions;
};
export type DecoratedStream = ReadableStream<Uint8Array> & {
    metadata?: Promise<unknown>;
    manifest?: Promise<Manifest>;
    header?: Header;
};
export type RewrapCacheOptions = {
    bypass?: boolean;
    maxAge?: number;
    pollInterval?: number;
};
export declare class RewrapCache {
    private cache?;
    private closer?;
    constructor(opts?: RewrapCacheOptions);
    get(key: Uint8Array): CryptoKey | undefined;
    set(key: Uint8Array, value: CryptoKey): void;
    close(): void;
}
export declare class OpenTDF {
    readonly policyEndpoint: string;
    readonly authProvider: AuthProvider;
    readonly dpopEnabled: boolean;
    defaultCreateOptions: Omit<CreateOptions, 'source'>;
    defaultReadOptions: Omit<ReadOptions, 'source'>;
    readonly dpopKeys: Promise<CryptoKeyPair>;
    private readonly rewrapCache;
    private tdf3Client;
    constructor({ authProvider, dpopKeys, defaultCreateOptions, defaultReadOptions, disableDPoP, policyEndpoint, rewrapCacheOptions, }: OpenTDFOptions);
    createNanoTDF(opts: CreateNanoTDFOptions): Promise<DecoratedStream>;
    /**
     * Creates a new collection object, which can be used to encrypt a series of data with the same policy.
     * @returns
     */
    createNanoTDFCollection(opts: CreateNanoTDFCollectionOptions): Promise<NanoTDFCollection>;
    createZTDF(opts: CreateZTDFOptions): Promise<DecoratedStream>;
    /**
     * Decrypts a nanotdf object. Optionally, stores the collection header and its DEK.
     * @param ciphertext
     */
    read(opts: ReadOptions): Promise<DecoratedStream>;
    close(): void;
}
export type NanoTDFCollection = {
    encrypt: (source: Source) => Promise<ReadableStream<Uint8Array>>;
    close: () => Promise<void>;
};
//# sourceMappingURL=opentdf.d.ts.map